#include "main.h"
#include "mqtt.h"
#include "frozen/frozen.h"

static struct channel_t s_channels[CHANNEL_MAX];
static int s_num_channels = 0;

static bool valid_gpio(const int gpio) {
  if (gpio == -1) {
    return true;
  }
  if (gpio < GPIO_MIN || gpio > GPIO_MAX) {
    return false;
  }
  return true;
}

int channel_get_total() {
  return s_num_channels;
}

bool channel_init(const char *fn) {
  char *            json;
  void *            h = NULL;
  struct json_token val;
  bool ret = false;

  char *name             = NULL;
  int   statusled        = -1;
  bool  statusled_invert = false;

  int idx;

  memset(s_channels, -1, sizeof(s_channels));
  s_num_channels = 0;

  json = json_fread(fn);
  if (!json) {
    LOG(LL_ERROR, ("%s: Could not json_fread()", fn));
    return false;
  }

  if (json_scanf(json, strlen(json), "{name:%Q, statusled:%d}", &name, &statusled) != 2) {
    LOG(LL_ERROR, ("Incomplete JSON: require 'name' and 'statusled' fields"));
    goto exit;
  }
  LOG(LL_INFO, ("Configuration: name='%s', statusled=%d", name, statusled));
  if (!valid_gpio(statusled)) {
    LOG(LL_ERROR, ("Status LED GPIO (%d) out of bounds [%d,%d]", statusled, GPIO_MIN, GPIO_MAX));
    goto exit;
  }
  json_scanf(json, strlen(json), "{statusled_invert:%B}", &statusled_invert);
  if (statusled_invert) {
    LOG(LL_INFO, ("Inverting statusled GPIO (%d)", statusled));
  }

  // Traverse Array
  while ((h = json_next_elem(json, strlen(json), h, ".channels", &idx, &val)) != NULL) {
    int led = -1, relay = -1, button = -1;

    LOG(LL_DEBUG, ("[%d]: [%.*s]", idx, val.len, val.ptr));
    if (val.len == 0 || !val.ptr) {
      continue;
    }
    if (json_scanf(val.ptr, val.len, "{led:%d, relay:%d, button:%d}", &led, &relay, &button) != 3) {
      LOG(LL_ERROR, ("Incomplete Channel JSON: require 'led' and 'relay' and 'button' fields"));
      goto exit;
    }
    LOG(LL_INFO, ("Channel[%d]: led=%d relay=%d button=%d ", idx, led, relay, button));
    if (!valid_gpio(led)) {
      LOG(LL_ERROR, ("LED GPIO (%d) out of bounds [%d,%d]", led, GPIO_MIN, GPIO_MAX));
      goto exit;
    }
    if (!valid_gpio(relay)) {
      LOG(LL_ERROR, ("Relay GPIO (%d) out of bounds [%d,%d]", relay, GPIO_MIN, GPIO_MAX));
      goto exit;
    }
    if (!valid_gpio(button)) {
      LOG(LL_ERROR, ("Button GPIO (%d) out of bounds [%d,%d]", button, GPIO_MIN, GPIO_MAX));
      goto exit;
    }
    if (idx == CHANNEL_MAX) {
      LOG(LL_ERROR, ("Too many channels (max is %d)", CHANNEL_MAX));
      goto exit;
    }
    s_channels[idx].button_gpio = button;
    s_channels[idx].relay_gpio  = relay;
    s_channels[idx].led_gpio    = led;
  }

  ret = true;
exit:
  if (ret) {
    if (statusled != -1) {
      statusled_init(statusled, statusled_invert);
    }

    s_num_channels = idx + 1;
    LOG(LL_INFO, ("Configuring %d channels", s_num_channels));

    for (; idx >= 0; idx--) {
      s_channels[idx].relay_state = 0;

      if (s_channels[idx].relay_gpio != GPIO_INVALID) {
        mgos_gpio_set_mode(s_channels[idx].relay_gpio, MGOS_GPIO_MODE_OUTPUT);
        mgos_gpio_write(s_channels[idx].relay_gpio, s_channels[idx].relay_state);
      }

      if (s_channels[idx].led_gpio != GPIO_INVALID) {
        mgos_gpio_set_mode(s_channels[idx].led_gpio, MGOS_GPIO_MODE_OUTPUT);
        mgos_gpio_write(s_channels[idx].led_gpio, s_channels[idx].relay_state);
      }

      if (s_channels[idx].button_gpio != GPIO_INVALID) {
        mgos_gpio_set_mode(s_channels[idx].button_gpio, MGOS_GPIO_MODE_INPUT);
        mgos_gpio_set_button_handler(s_channels[idx].button_gpio, MGOS_GPIO_PULL_UP, MGOS_GPIO_INT_EDGE_NEG, 20, channel_handler, (void*) 1);
      }
    }
  }
  if (name) {
    free(name);
  }
  if (json) {
    free(json);
  }
  return ret;
}

uint8_t channel_gpio_by_idx(int idx) {
  if (idx < 0 || idx >= s_num_channels) {
    return GPIO_INVALID;
  }
  return s_channels[idx].button_gpio;
}

uint8_t channel_idx_by_gpio(int gpio) {
  uint8_t i;

  for (i = 0; i < channel_get_total(); i++) {
    if (gpio == s_channels[i].button_gpio) {
      return i;
    }
  }
  return GPIO_INVALID;
}

void channel_set(int idx, bool state) {
  double now = mg_time();

  if (idx < 0 || idx >= channel_get_total()) {
    return;
  }

  s_channels[idx].button_last_change = now;
  s_channels[idx].relay_state        = state;
  if (s_channels[idx].relay_gpio != GPIO_INVALID) {
    mgos_gpio_write(s_channels[idx].relay_gpio, state);
  }
  if (s_channels[idx].led_gpio != GPIO_INVALID) {
    mgos_gpio_write(s_channels[idx].led_gpio, state);
  }

  mqtt_publish_stat("channel", "{idx: %d, relay_state: %d}", idx, channel_get(idx));
}

bool channel_get(int idx) {
  if (idx < 0 || idx >= channel_get_total()) {
    return false;
  }

  return s_channels[idx].relay_state == 1;
}

// Channel handler arg is non-NULL if it was a pushbutton, and NULL if it was an RPC.
void channel_handler(int gpio, void *arg) {
  uint8_t idx;
  bool    state;

  idx = channel_idx_by_gpio(gpio);
  if (idx == GPIO_INVALID) {
    LOG(LL_ERROR, ("GPIO %d is not a valid button", gpio));
    return;
  }

  if (!arg) { // pushbutton
    LOG(LL_INFO, ("RPC triggered channel %d", idx));
  } else {
    LOG(LL_INFO, ("Button on GPIO %d triggered channel %d", gpio, idx));
  }

  state = channel_get(idx);
  channel_set(idx, !state);

  (void)arg;
}
